@@ -2,8 +2,8 @@
 
 about = {
     "name": "softchat",
-    "version": "1.0",
-    "date": "2021-02-22",
+    "version": "1.2",
+    "date": "2021-04-01",
     "description": "convert twitch/youtube chat into softsubs",
     "author": "ed",
     "license": "MIT",
@@ -16,6 +16,8 @@ import sys
 import time
 import json
 import zlib
+import shlex
+import string
 import base64
 import random
 import requests
@@ -25,11 +27,15 @@ import shutil
 import argparse
 import tempfile
 import colorsys
+import queue
+import multiprocessing
 import subprocess as sp
-from PIL import ImageFont, ImageDraw, Image
+from PIL import Image
 from .util import debug, info, warn, error, init_logger
-from .util import WINDOWS
-from .util import shell_esc, zopen
+from .util import WINDOWS, HAVE_FONTFORGE
+from .util import shell_esc, zopen, tt, hms, get_ff_dur, load_fugashi
+from .mproc import TextStuff, gen_msg_thr, gen_msg_initializer
+from .ass import assan, segment_msg, render_msegs
 
 
 try:
@@ -43,88 +49,6 @@ except ImportError:
 if __name__ == "__main__":
     init_logger("-d" in sys.argv)
 
-message_translation_table = "".maketrans(
-    "",
-    "",
-    # Skin tone modifiers do not render in subtitles.
-    "🏻🏼🏽🏾🏿",
-)
-
-
-try:
-    # help python find libmecab.dll, adjust this to fit your env if necessary
-    dll_path = None
-    for base in sys.path:
-        x = os.path.join(base, "fugashi")
-        if os.path.exists(os.path.join(x, "cli.py")) and not dll_path:
-            dll_path = x
-        x2 = os.path.join(x, "../../../lib/site-packages/fugashi")
-        if os.path.exists(x2):
-            dll_path = x2
-            break
-
-    if not dll_path:
-        raise Exception("could not find fugashi installation path")
-
-    if WINDOWS:
-        os.add_dll_directory(dll_path)
-
-    from fugashi import Tagger
-
-    dicrc = os.path.join(dll_path, "dicrc")
-    with open(dicrc, "wb") as f:
-        f.write(
-            "\n".join(
-                [
-                    r"node-format-yomi = %f[9] ",
-                    r"unk-format-yomi = %m",
-                    r"eos-format-yomi  = \n",
-                    "",
-                ]
-            ).encode("utf-8")
-        )
-
-    wakati = Tagger("-Owakati")
-    yomi = Tagger("-Oyomi -r " + dicrc.replace("\\", "\\\\"))
-
-    # import MeCab
-    # wakati = MeCab.Tagger('-Owakati')
-    HAVE_TOKENIZER = True
-    info("found fugashi")
-except:
-    HAVE_TOKENIZER = False
-    import traceback
-
-    warn("could not load fugashi:\n" + traceback.format_exc() + "-" * 72 + "\n")
-
-
-def find_fontforge():
-    """fontforge comes with a full python env on windows (nice), try to find it"""
-    search_dirs = []
-    for k in ["ProgramFiles", "ProgramFiles(x86)", "ProgramW6432"]:
-        v = os.environ.get(k)
-        if v and v not in search_dirs:
-            search_dirs.append(v)
-
-    for sdir in search_dirs:
-        ffpy = os.path.join(sdir, "FontForgeBuilds/bin/ffpython.exe")
-        if os.path.exists(ffpy):
-            return ffpy
-
-    # actually they do on macos as well
-    ffpy = "/Applications/FontForge.app/Contents/MacOS/FFPython"
-    if os.path.exists(ffpy):
-        return ffpy
-
-    return None
-
-
-HAVE_FONTFORGE = True
-try:
-    import fontforge
-except:
-    HAVE_FONTFORGE = find_fontforge()
-
 
 HAVE_MAGICK = False
 try:
@@ -138,321 +62,6 @@ except:
     pass
 
 
-class TextStuff(object):
-    def __init__(self, sz, fontdir, emote_scale):
-        self.sz = sz
-
-        if fontdir:
-            fontdir = fontdir.rstrip(os.sep)
-            if fontdir.endswith("noto-hinted"):
-                fontdir = fontdir.rsplit(os.sep, 1)[0]
-
-        src = os.path.join("noto-hinted", "NotoSansCJKjp-Regular.otf")
-        ok, self.otf_src = self.resolve_path(src, fontdir)
-        if not ok:
-            err = "could not locate fonts after trying these locations:"
-            raise Exception("\n  ".join([err] + self.otf_src))
-
-        dst = os.path.join("noto-hinted", "SquishedNotoSansCJKjp-Regular.otf")
-        ok, self.otf_mod = self.resolve_path(dst, fontdir, self.otf_src)
-        if ok:
-            info(f"found font: {self.otf_mod}")
-        else:
-            self.conv_otf()
-
-        # mpv=870 pil=960 mul=0.9 (due to fontsquish?)
-        self.font = ImageFont.truetype(self.otf_mod, size=int(sz * 0.9 + 0.9))
-        self.im = Image.new("RGB", (3840, 2160), "white")
-        self.imd = ImageDraw.Draw(self.im)
-        self.pipe_width = self.font.getsize("|")[0]
-        self.font_ofs = self.font.getmetrics()[1]
-        self.cache = {}
-        self.vsize = self.caching_vsize
-        self.emote_scale = emote_scale
-        self.emote_repl = "/%"  # good enough
-        self.emote_vsz = self.vsize_impl(self.emote_repl, False)
-
-    def resolve_path(self, path, suggested, refpath=None):
-        # wow really didn't think this through
-        home = os.path.expanduser("~")
-        basedirs = [
-            os.path.dirname(os.path.realpath(__file__)),
-            os.getcwd(),
-            home,
-            os.path.join(home, "fonts"),
-            os.path.join(home, ".fonts"),
-            os.path.join(home, "Downloads"),
-            tempfile.gettempdir(),
-        ]
-
-        if suggested:
-            basedirs = [suggested] + basedirs
-
-        ret = [os.path.join(x, path) for x in basedirs]
-
-        if refpath:
-            # prefer same directory as the other file
-            for x in ret:
-                if x.rsplit(os.sep, 1)[0] == refpath.rsplit(os.sep, 1)[0]:
-                    ret = [x] + [y for y in ret if y != x]
-                    break
-
-        for fp in ret:
-            try:
-                if os.path.exists(fp):
-                    return True, fp
-            except:
-                pass
-
-        return False, ret
-
-    def conv_otf(self):
-        from fontTools.ttLib import TTFont, newTable
-
-        info("creating squished font, pls wait")
-        mul = 1.1
-        font = TTFont(self.otf_src)
-        baseAsc = font["OS/2"].sTypoAscender
-        baseDesc = font["OS/2"].sTypoDescender
-        font["hhea"].ascent = round(baseAsc * mul)
-        font["hhea"].descent = round(baseDesc * mul)
-        font["OS/2"].usWinAscent = round(baseAsc * mul)
-        font["OS/2"].usWinDescent = round(baseDesc * mul) * -1
-
-        xfn = "softchat.xml"
-        font.saveXML(xfn, tables=["name"])
-        with open(xfn, "r", encoding="utf-8") as f:
-            xml = f.read()
-
-        xml = xml.replace("Noto Sans", "Squished Noto Sans")
-        xml = xml.replace("NotoSans", "SquishedNotoSans")
-        with open(xfn, "w", encoding="utf-8") as f:
-            f.write(xml)
-
-        font["name"] = newTable("name")
-        font.importXML(xfn)
-        os.unlink(xfn)
-
-        try:
-            del font["post"].mapping["Delta#1"]
-        except:
-            pass
-
-        for fp in self.otf_mod:
-            try:
-                fdir = fp.rsplit(os.sep, 1)[0]
-                os.makedirs(fdir, exist_ok=True)
-                with open(fp, "wb") as f:
-                    f.write(b"h")
-
-                os.unlink(fp)
-                break
-            except:
-                fp = None
-
-        if not fp:
-            err = "could not write modified font to any of the following locations:"
-            raise Exception("\n  ".join([err] + self.otf_mod))
-
-        self.otf_mod = fp
-        info(f"writing {self.otf_mod}")
-        font.save(self.otf_mod)
-
-    def unemote(self, text):
-        return "".join(
-            [
-                c if u < 0xE000 or u > 0xF8FF else self.emote_repl
-                for c, u in zip(text, [ord(x) for x in text])
-            ]
-        )
-
-    def vsize_impl(self, text, msg_emotes):
-        if msg_emotes:
-            text = self.unemote(text)
-
-        w, h = self.imd.textsize("|" + text.replace("\n", "\n|"), self.font)
-
-        if msg_emotes and self.emote_scale > 1.01:
-            em_dw, em_dh = [x * self.emote_scale - x for x in self.emote_vsz]
-            add_w = 0
-            add_h = 0
-            for ln in text.split("\n"):
-                # yolo: self.emote_repl unlikely to be in messages
-                num_emotes = min(len(msg_emotes), ln.count(self.emote_repl))
-                if num_emotes:
-                    ln_w = em_dw * num_emotes
-                    add_w = max(add_w, ln_w)
-                    add_h += em_dh
-            w += add_w
-            h += add_h
-
-        return w - self.pipe_width, h
-
-    def caching_vsize(self, text, msg_emotes):
-        if len(text) > 24:
-            return self.vsize_impl(text, msg_emotes)
-
-        # faster than try/catch and get(text,None)
-        if text in self.cache:
-            return self.cache[text]
-
-        ret = self.vsize_impl(text, msg_emotes)
-        self.cache[text] = ret
-        return ret
-
-    def unrag(self, text, width, msg_emotes):
-        """
-        copied from  http://xxyxyz.org/line-breaking/
-        license clarified per email:
-          BSD or MIT since the EU doesn't recognize Public Domain
-
-        only change so far is replacing len(w) with vsize(w+"_")[0]
-        """
-        words = text.split()
-        count = len(words)
-        offsets = [0]
-        for w in words:
-            offsets.append(offsets[-1] + self.vsize(w + "_", msg_emotes)[0])
-
-        minima = [0] + [10 ** 20] * count
-        breaks = [0] * (count + 1)
-        for i in range(count):
-            j = i + 1
-            while j <= count:
-                w = offsets[j] - offsets[i] + j - i - 1
-                if w > width:
-                    break
-                cost = minima[i] + (width - w) ** 2
-                if cost < minima[j]:
-                    minima[j] = cost
-                    breaks[j] = i
-                j += 1
-
-        lines = []
-        j = count
-        while j > 0:
-            i = breaks[j]
-            lines.append(" ".join(words[i:j]))
-            j = i
-        lines.reverse()
-        return lines
-
-
-def hms(s):
-    h, s = divmod(s, 3600)
-    m, s = divmod(s, 60)
-    return "{:d}:{:02d}:{:05.2f}".format(int(h), int(m), s)
-
-
-def tt(s):
-    s = int(s)
-
-    h, s = divmod(s, 3600)
-    m, s = divmod(s, 60)
-    if h > 0:
-        return "{:d}:{:02d}:{:02d}".format(h, m, s)
-    return "{:d}:{:02d}".format(m, s)
-
-
-def assan(x):
-    # there is no standardization on escaping ["{", "}", "\\"]:
-    #   the commented one is for aegisub,
-    #   the enabled one is for mpv
-
-    # aegsiub:
-    # return x.replace("{", "<").replace("}", ">").replace('\\', '\\{}')
-
-    # mpv:
-    ret = ""
-    for c, nc in zip(x, x[1:] + "\n"):
-        if c == "{":
-            ret += "\\{"
-        elif c == "}":
-            ret += "\\}"
-        elif c == "\\" and nc in ["N", "n", "h"]:
-            ret += "\\\\"
-        else:
-            ret += c
-
-    # mpv:
-    #   there is no way to encode a literal \ before a markup {
-    #   so the lines must end with a whitespace
-    return ret + " "
-
-
-def get_ff_dur(fn):
-    # ffprobe -hide_banner -v error -select_streams v:0 -show_entries stream=duration -of default=noprint_wrappers=1:nokey=1 foo.mkv
-    #   webm: n/a
-    #   mkv: n/a
-    #   mp4: numSec
-    # ffprobe -hide_banner -v error -show_entries format=duration -of default=noprint_wrappers=1:nokey=1 foo.mkv
-    #   webm: numSec
-    #   mkv: numSec
-    #   mp4: numSec (0.06 sec larger)
-
-    # Try to grab the duration for the video stream, if available, otherwise fall back to format duration (webm)
-
-    if fn.lower().endswith("mkv"):
-        ret = sp.check_output(
-            [
-                "ffprobe",
-                "-hide_banner",
-                "-v",
-                "error",
-                "-select_streams",
-                "v:0",
-                "-show_entries",
-                "stream_tags=duration",
-                "-show_entries",
-                "format=duration",
-                "-of",
-                "default=noprint_wrappers=1:nokey=1",
-                fn,
-            ]
-        )
-        r = ret.decode("utf-8").strip().split("\n")
-
-        if len(r) == 2 and ":" in r[0]:
-            h, m, s = [float(x) for x in r[0].split(":")]
-            return 60 * (60 * h + m) + s
-        else:
-            return float(r[-1])
-    elif fn.lower().endswith("mp4"):
-        ret = sp.check_output(
-            [
-                "ffprobe",
-                "-hide_banner",
-                "-v",
-                "error",
-                "-select_streams",
-                "v:0",
-                "-show_entries",
-                "stream=duration",
-                "-of",
-                "default=noprint_wrappers=1:nokey=1",
-                fn,
-            ]
-        )
-        return float(ret.split(b"\n", 1)[0])
-    elif fn.lower().endswith("webm"):
-        ret = sp.check_output(
-            [
-                "ffprobe",
-                "-hide_banner",
-                "-v",
-                "error",
-                "-show_entries",
-                "format=duration",
-                "-of",
-                "default=noprint_wrappers=1:nokey=1",
-                fn,
-            ]
-        )
-        return float(ret)
-    else:
-        raise Exception(f"Unknown video format {fn}")
-
-
 def convert_old(m):
     o = {
         "action_type": "add_chat_item",
@@ -562,6 +171,7 @@ def cache_emotes(emotes, emote_dir, overwrite):
                 # that are too high will destroy detail.
                 # There is no single best option for all emotes.
                 "-contrast-stretch", "3%x9%",
+                "-channel", "rgb",
                 "-negate",
                 fname,
             ])
@@ -639,6 +249,7 @@ def main():
     ap.add_argument("-m", metavar="MODE", type=int, default=1, help="mode, 1=box, 2=danmaku")
     ap.add_argument("-r", metavar="WxH", type=str, default="1280x720", help="video res")
     ap.add_argument("-b", metavar="WxH+X+Y", type=str, default=None, help="subtitle area")
+    ap.add_argument("-j", metavar="CORES", type=int, default=0, help="number of cores to use (0=all, 1=single-threaded)")
     ap.add_argument("--sz", metavar="POINTS", type=int, default=0, help="font size")
     ap.add_argument("--spd", metavar="SPEED", type=int, default=256, help="[danmaku] pixels/sec")
     ap.add_argument("--spread", action="store_true", help="[danmaku] even distribution")
@@ -660,11 +271,13 @@ def main():
     ap.add_argument("--emote_nofont", action="store_true", help="[DEBUG] disable font generation")
     ap.add_argument("--emote_install", action="store_true", help="install emote fonts into media player folders")
     ap.add_argument("--emote_install_dir", type=str, default=None, help="Optional directory to install fonts, if not present will try to determine a default system location.")
+    ap.add_argument("--no_errdep_emotes", action="store_true", help="ignore missing dependencies for requested emote stuff; disable the unsatisfiable arguments and continue")
     ap.add_argument("fn", metavar="JSON_FILE", nargs="+")
     ar = ap.parse_args()
     # fmt: on
 
-    if ar.kana and not HAVE_TOKENIZER:
+    have_fugashi = bool(load_fugashi(write_cfg=True))
+    if ar.kana and not have_fugashi:
         error("you requested --kana but mecab failed to load")
         sys.exit(1)
 
@@ -679,7 +292,39 @@ def main():
         if err:
             err = ", ".join(err)
             error(f"you requested --emote_font but {err} is not installed")
-            sys.exit(1)
+            if ar.no_errdep_emotes:
+                ar.emote_font = None
+                ar.emote_sz = 1
+            else:
+                sys.exit(1)
+
+    media_fn = None
+    if ar.media and os.path.isfile(ar.media):
+        media_fn = ar.media
+    else:
+        for ext in ["webm", "mp4", "mkv"]:
+            f = ar.fn[0]
+            while not media_fn and "." in f:
+                f = f.rsplit(".", 1)[0]
+                mfn = f + "." + ext
+                if os.path.isfile(mfn):
+                    media_fn = mfn
+                    break
+
+                mfn = os.path.join(os.getcwd(), os.path.split(mfn)[1])
+                if os.path.isfile(mfn):
+                    media_fn = mfn
+                    break
+
+    base_fn = ar.fn[0].rsplit(".json", 1)[0]
+    if media_fn:
+        base_fn = media_fn.rsplit(".", 1)[0]
+
+    out_fn = base_fn + ".ass"
+    font_fn = base_fn + ".ttf"
+    if WINDOWS:
+        allowed = string.ascii_letters + string.digits + ".,-"
+        font_fn = "".join([x if x in allowed else "_" for x in font_fn])
 
     emote_dir = "emotes"
     if ar.emote_cache:
@@ -815,7 +460,6 @@ def main():
         ar.emote_font = False
 
     font_name = "Squished Noto Sans CJK JP Regular"
-    font_fn = ar.fn[0].rsplit(".", 1)[0] + ".ttf"
     emote_shortcuts = dict()
     filled_emotes = []
     if ar.emote_font:
@@ -823,7 +467,7 @@ def main():
         cache_emotes(emotes, emote_dir, ar.emote_refilter)
 
         # Try to avoid collisions if someone does install these as system fonts.
-        font_hash = hashlib.sha512(ar.fn[0].encode("utf-8")).digest()
+        font_hash = hashlib.sha512(base_fn.encode("utf-8")).digest()
         font_hash = base64.urlsafe_b64encode(font_hash)[:16].decode("ascii")
         font_name = f"SoftChat Custom Emotes {font_hash}"
         emote_shortcuts, filled_emotes = generate_font(
@@ -962,19 +606,6 @@ def main():
         info(f"Dropping {len(droplist)} duplicate chat entries within threshold")
     jd = [m for m in jd if m["message_id"] not in droplist]
 
-    media_fn = None
-    if ar.media and os.path.isfile(ar.media):
-        media_fn = ar.media
-    else:
-        for ext in ["webm", "mp4", "mkv"]:
-            f = ar.fn[0]
-            while not media_fn and "." in f:
-                f = f.rsplit(".", 1)[0]
-                mfn = f + "." + ext
-                if os.path.isfile(mfn):
-                    media_fn = mfn
-                    break
-
     cdur_msg = None
     cdur_err = "could not verify chat duration"
     v_dur = None
@@ -1007,12 +638,6 @@ def main():
     else:
         info(cdur_msg)
 
-    ptn_kanji = re.compile(r"[\u4E00-\u9FAF]")
-    ptn_kana = re.compile(r"[\u3040-\u30FF]")
-    ptn_ascii = re.compile(r"[a-zA-Z]")
-    ptn_pre = re.compile(r"([　、。〇〉》」』】〕〗〙〛〜〞〟・…⋯！＂）＊－．／＞？＠＼］＿～｡｣･￭￮]+)")
-    ptn_post = re.compile(r"([〈《「『【〔〖〘〚〝（＜［｀｢]+)")
-
     info(f"deduping nicknames in {len(jd)} chat entries")
     pair_seen = set()
     nick_dupes = set()
@@ -1049,131 +674,21 @@ def main():
 
     msgs = []
     info("converting")
-    for n_msg, msg in enumerate(jd):
-        txt = msg.get("message", "") or ""
-        txt = txt.translate(message_translation_table)
-        if "amount" not in msg and "money" not in msg and txt == "":
-            txt = "--"
-
-        t_fsec = msg["time_in_seconds"]
-        t_isec = int(t_fsec)
-        t_hms = msg["time_text"]
-
-        if t_hms.startswith("-") or t_isec < 0 or t_isec > 4096 * 4096:
-            continue
-
-        # time integrity check
-        try:
-            h, m, s = [int(x) for x in t_hms.split(":")]
-            t_isec2 = 60 * (60 * h + m) + s
-        except:
-            m, s = [int(x) for x in t_hms.split(":")]
-            t_isec2 = 60 * m + s
-
-        if int(t_fsec) != t_isec or t_isec != t_isec2:
-            # at some point, someone observed a difference between time_text
-            # and time_in_seconds, so please let me have a look at
-            # your chatlog json if you end up here
-            raise Exception(
-                f"time drift [{t_fsec}] [{t_isec}] [{t_isec2}] [{t_hms}]\n  (pls provide this chat-rip to ed)"
-            )
-
-        msg_emotes = []
-        if ":" in txt and ar.emote_font:
-            old_txt = txt
-            for k, v in emote_shortcuts.items():
-                txt = txt.replace(k, v)
-
-            if txt != old_txt:
-                msg_emotes = [u for u in txt if ord(u) >= 0xE000 and ord(u) <= 0xF8FF]
-
-        # wordwrap gets wonky when emotes are 2big
-        # so ensure whtiespace between text and emote regions
-        if msg_emotes and ar.emote_sz >= 1.5:
-            txt2 = ""
-            was_emote = False
-            for c, u in zip(txt, [ord(x) for x in txt]):
-                if u >= 0xE000 and u <= 0xF8FF:
-                    if not was_emote and txt2:
-                        txt2 += " "
-                    was_emote = True
-                else:
-                    if was_emote:
-                        was_emote = False
-                        txt2 += " "
-                txt2 += c
-            txt = txt2
-
-        n_ascii = len(ptn_ascii.findall(txt))
-        n_kanji = len(ptn_kanji.findall(txt))
-        n_kana = len(ptn_kana.findall(txt))
-
-        # if the amount of ascii compared to kanji/kana
-        # is less than 30%, assume we'll need MeCab
-        is_jp = (n_kanji + n_kana) / (n_kanji + n_kana + n_ascii + 0.1) > 0.7
-
-        # transcription from kanji to kana if requested
-        if ar.kana and is_jp and n_kanji:
-            txt2 = yomi.parse(txt)
-            txt = ""
-            for ch in txt2:
-                co = ord(ch)
-                # hiragana >= 0x3041 and <= 0x3096
-                # katakana >= 0x30a1 and <= 0x30f6
-                if co >= 0x30A1 and co <= 0x30F6:
-                    txt += chr(co - (0x30A1 - 0x3041))
-                else:
-                    txt += ch
-
-        if ar.m == 1:
-            wrap_width = bw
-        else:
-            # vsfilter wraps it anyways orz
-            wrap_width = vw / 2
-
-        if n_msg % 100 == 0 and len(z.cache) > 1024 * 64:
-            z.cache = {}
-
-        # wrap to specified width
-        # by splitting on ascii whitespace
-        vtxt = z.unrag(txt, wrap_width, msg_emotes)
-        vsz = z.vsize("\n".join(vtxt), msg_emotes)
-
-        if vsz[0] >= bw and is_jp:
-            # too wide, is japanese,
-            # try wrapping on japanese punctuation instead
-            vtxt = ptn_pre.sub("\\1\n", txt)
-            vtxt = ptn_post.sub("\n\\1", vtxt)
-            vtxt = vtxt.split("\n")
-            vsz = z.vsize("\n".join(vtxt), msg_emotes)
-
-            if vsz[0] >= bw and HAVE_TOKENIZER:
-                # still too wide, wrap on word-boundaries
-                vtxt = z.unrag(wakati.parse(txt), bw, msg_emotes)
-                vtxt = [x.replace(" ", "") for x in vtxt]
-
-                for n in range(1, len(vtxt)):
-                    # move most punctuation to the prev line
-                    ln = vtxt[n]
-                    m = ptn_pre.search(ln)
-                    if m and m.start() == 0:
-                        vtxt[n - 1] += ln[: m.end()]
-                        vtxt[n] = ln[m.end() :]
-
-                vsz = z.vsize("\n".join(vtxt), msg_emotes)
-
-        vtxt = [x for x in vtxt if x.strip()]
-
+    conv_t0 = time.time()
+    for n_msg, msg, vtxt, vsz, t_fsec, t_hms, msg_emotes in gen_msgs(
+        jd, vw, bw, ar, emote_shortcuts, have_fugashi
+    ):
         # pillow height calculation is off by a bit; this is roughly it i think
         sx, sy = vsz
         sy = int(sy + fofs * 2.5 + 0.8)
 
         if n_msg % 1000 == 0:
             info(
-                "  {} / {}   {}%   {}   {}\n   {}\n".format(
+                "  {} / {}   {}%   {}/s   {}   {}\n   {}  \n".format(
                     n_msg,
                     len(jd),
                     int((n_msg * 100) / len(jd)),
+                    int(n_msg / (time.time() - conv_t0)),
                     t_hms,
                     [sx, sy],
                     "\n   ".join(vtxt),
@@ -1216,17 +731,14 @@ def main():
         #    break
 
     vis = []
-    if ar.fn[0].lower().endswith(".json"):
-        out_fn = ar.fn[0][:-5] + ".ass"
-    else:
-        out_fn = ar.fn[0] + ".ass"
 
     info(f"creating {out_fn}")
     with open(out_fn, "wb") as f:
         f.write(
             """\
 [Script Info]
-Title: https://ocv.me/dev/?softchat.py
+Title: https://github.com/9001/softchat
+; {cmd}
 ScriptType: v4.00+
 WrapStyle: 0
 ScaledBorderAndShadow: yes
@@ -1247,7 +759,11 @@ Style: a,{font},{sz},&H00FFFFFF,&H000000FF,&H00000000,&H80000000,0,0,0,0,100,100
 [Events]
 Format: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
 """.format(
-                vw=vw, vh=vh, sz=ar.sz, font=font_name
+                vw=vw,
+                vh=vh,
+                sz=ar.sz,
+                font=font_name,
+                cmd=" ".join(map(shlex.quote, sys.argv[1:])),
             ).encode(
                 "utf-8"
             )
@@ -1531,43 +1047,13 @@ Format: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
                                 best = this
                                 y = y1 - h
 
-                if msg_emotes and (ar.emote_fill or filled_emotes):
-                    # txt = rf"{{\clip(1,m 0 0 l 1000 0 1000 1000 0 1000)}}" + txt
-                    # txt = rf"{{\clip(0,0,700,700)}}" + txt
-                    # you can't \move a \clip sick there goes that idea
-                    txt2 = ""
-                    emotes = ""
-                    for c, u in list(zip(txt, [ord(x) for x in txt])) + [[None, 0]]:
-                        is_emote = u >= 0xE000 and u <= 0xF8FF
-                        if is_emote and (ar.emote_fill or c in filled_emotes):
-                            emotes += c
-                        else:
-                            if emotes:
-                                # aegisub @fs200/emotefont:
-                                #   {\pos(620,500)}|{\c&H660000&\fscx442\fsp-220}{\fscx100\fsp0\c}龖龖{\c&H0000FF&\1a&HBF&}龖龖龖|
-                                #   {\pos(620,720)}|龖龖龖龖龖|
-
-                                # emote font size
-                                fsz = ar.sz
-                                if ar.emote_sz:
-                                    esz = fsz * ar.emote_sz
-                                    sz1 = f"\\fs{esz}"
-                                    sz2 = f"\\fs{fsz}"
-                                else:
-                                    esz = fsz
-                                    sz1 = sz2 = ""
-
-                                # compensate the 1100 padding in fff by subtracting a constant,
-                                scx = int(len(emotes) * 110) - 10  # ~109 otherwise
-
-                                # and bump this a bit to shift some of the padding to the left
-                                fsp = esz / 0.9111  # 0.9091
-
-                                txt2 += f"{{\\c&H{bgr_msg}\\fscx{scx}\\fsp-{fsp}{sz1}}}\ue000{{\\fscx100\\fsp0\\c&H{bgr_fg}\\1a&H00&\\bord1\\shad0}}{emotes}{{\\bord{bord}\\shad{shad}{sz2}}}"
-                                emotes = ""
-                            if u:
-                                txt2 += c
-                    txt = txt2
+                if msg_emotes and (
+                    ar.emote_fill or filled_emotes or ar.emote_sz > 1.01
+                ):
+                    msegs = segment_msg(txt, ar.emote_fill, filled_emotes)
+                    txt = render_msegs(
+                        msegs, ar.sz, ar.sz * ar.emote_sz, bgr_msg, bgr_fg, bord, shad
+                    )
 
                 y = int(y)
                 txt = rf"{{\move({vw},{y+h},{-w},{y+h})\3c&H{bgr_nick}&}}{txt}{{\fscx40\fscy40\bord1}}\N{nick}"
@@ -1684,6 +1170,23 @@ Format: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
     info(f"finished in {t1_main-t0_main:.2f} sec")
 
 
+def gen_msgs(jd, vw, bw, ar, emote_shortcuts, have_fugashi):
+    j = ar.j
+    if j == 0:
+        j = os.cpu_count()
+
+    initargs = [gen_msg_thr, ar, vw, bw, emote_shortcuts, have_fugashi]
+
+    with multiprocessing.Pool(
+        j, initializer=gen_msg_initializer, initargs=initargs
+    ) as pool:
+        # Cannot return the generator directly, since the context manager will close the pool
+        # 100 was picked experimentally and seems to perform well for both 4c8t and 16c/32t.
+        for x in pool.imap(gen_msg_thr, enumerate(jd), 100):
+            if x:
+                yield x
+
+
 if __name__ == "__main__":
     main()
 
